package net.jaekl.squelch;

import static org.junit.Assert.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;

import net.jaekl.squelch.util.FileMock;
import net.jaekl.squelch.util.FileUtilMock;

import org.junit.Test;

public class ArgsTest 
{
	private static class Result {
		private boolean m_retVal;
		private Args m_args;
		private String m_stdout;
		
		public Result(boolean retVal, Args args, String stdout) {
			m_retVal = retVal;
			m_args = args;
			m_stdout = stdout;
		}
		
		public boolean getRetVal() { return m_retVal; }
		public Args getArgs() { return m_args; }
		public String getStdout() { return m_stdout; }
	}

	@Test
	public void test_variousParams() throws IOException 
	{
		Result result = null;
		
		String[] noParams = {};
		result = runWithParams(noParams);
		assertTrue(result.getRetVal());
		assertEquals(false, result.getArgs().isHelp());
		assertEquals(Args.DEFAULT_JDBC_URL, result.getArgs().getUrl());
		assertEquals("", result.getArgs().getUser());
		assertEquals("", result.getArgs().getPass());
		assertEquals("", result.getStdout());
		
		String[] helpOnly = { "--help" };
		result = runWithParams(helpOnly);
		assertTrue(result.getRetVal());
		assertEquals(true, result.getArgs().isHelp());
		assertEquals(Args.DEFAULT_JDBC_URL, result.getArgs().getUrl());
		assertEquals("", result.getArgs().getUser());
		assertEquals("", result.getArgs().getPass());
		validateUsageMessage(result.getStdout());
		
		String[] oneTwoThree = { "-r", "one", "-u", "two", "-p", "three" };
		result = runWithParams(oneTwoThree);
		assertTrue(result.getRetVal());
		assertEquals(false, result.getArgs().isHelp());
		assertEquals("one", result.getArgs().getUrl());
		assertEquals("two", result.getArgs().getUser());
		assertEquals("three", result.getArgs().getPass());
		assertEquals("", result.getStdout());
		
		String[] invalid = { "--bogus" };
		result = runWithParams(invalid);
		assertFalse(result.getRetVal());
		assertEquals(false, result.getArgs().isHelp());
		assertEquals(Args.DEFAULT_JDBC_URL, result.getArgs().getUrl());
		assertEquals("", result.getArgs().getUser());
		assertEquals("", result.getArgs().getPass());
		validateUsageMessage(result.getStdout());
	}
	
	@Test
	public void test_configFile_success() throws IOException
	{
		final String CONFIG_FILE_NAME = "configFileName";
		final String SIMPLE_CONTENT = "db.user=fred\n"
				                    + "db.password=p455w0rd\n"
				                    + "db.url=jdbc:postgresql://localhost/cfb\n";
		
		Result result = null;
		
		FileUtilMock.mock_setInst();
		FileMock configFileMock = FileUtilMock.mock_getInst().mock_newFile(CONFIG_FILE_NAME);
		FileUtilMock.mock_getInst().mock_setReadable(configFileMock, true);
		configFileMock.mock_setContent(SIMPLE_CONTENT.getBytes(StandardCharsets.UTF_8));
		
		String[] simple = { "-c", CONFIG_FILE_NAME };
		result = runWithParams(simple);
		assertTrue(result.getRetVal());
		assertEquals(false, result.getArgs().isHelp());
		assertEquals("jdbc:postgresql://localhost/cfb", result.getArgs().getUrl());
		assertEquals("fred", result.getArgs().getUser());
		assertEquals("p455w0rd", result.getArgs().getPass());
		assertEquals("", result.getStdout());
	}
	
	@Test
	public void test_configFile_notReadable()
	{
		final String CONFIG_FILE_NAME = "configFileName";
		
		FileUtilMock.mock_setInst();
		FileMock configFileMock = FileUtilMock.mock_getInst().mock_newFile(CONFIG_FILE_NAME);
		FileUtilMock.mock_getInst().mock_setReadable(configFileMock, false);
		
		String[] params = { "-c", CONFIG_FILE_NAME };
		try {
			runWithParams(params);
			fail("Should have thrown an IOException");
		}
		catch (IOException exc) {
			// This is the success path
		}
	}
	
	// We expect the stdout string to contain a usage message.
	// Do some sanity-checking to ensure that it looks vaguely correct.
	private void validateUsageMessage(String stdout) 
	{
		assertTrue(stdout.contains(this.getClass().getName()));
		assertTrue(stdout.contains("--help"));
	}

	private Result runWithParams(String[] params)
		throws IOException
	{
		try (
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				PrintWriter pw = new PrintWriter(new OutputStreamWriter(baos, StandardCharsets.UTF_8));
			)
		{
			Args args = new Args(this.getClass().getName());
			boolean retVal = args.parseArgs(pw, params);

			pw.flush();
			baos.flush();
			String stdout = baos.toString(StandardCharsets.UTF_8.name());
			
			return new Result(retVal, args, stdout);
		}		
	}
}
