package net.jaekl.squelch.stmt;

import static org.junit.Assert.assertEquals;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import java.sql.Types;

import net.jaekl.squelch.sql.Column;
import net.jaekl.squelch.sql.Row;

import org.junit.Test;

public class TabularTest {
	@Test
	public void test_centrePad() {
		Tabular tabular = new TabularMock();
		
		assertEquals("Vestibule", tabular.centrePad("Vestibule", 2));
		assertEquals("  Fred  ", tabular.centrePad("Fred", 8));
		assertEquals("NULL", tabular.centrePad(null, 0));
		assertEquals("  NULL  ", tabular.centrePad(null, 8));
		assertEquals("   Wilma    ", tabular.centrePad("Wilma", 12));
	}

	@Test
	public void test_classForSqlType() {
		Object[][] data = { 
				{ Types.ARRAY, Object.class },
				{ Types.BIGINT, Object.class },
				{ Types.BINARY, Object.class },
				{ Types.BIT, Object.class },
				{ Types.BLOB, Object.class },
				{ Types.BOOLEAN, Boolean.class },
				{ Types.CHAR, Character.class },
				{ Types.CLOB, Object.class },
				{ Types.DATALINK, Object.class },
				{ Types.DATE, java.util.Date.class },
				{ Types.DECIMAL, Double.class },
				{ Types.DISTINCT, Object.class },
				{ Types.DOUBLE, Double.class },
				{ Types.FLOAT, Double.class },
				{ Types.INTEGER, Long.class },
				{ Types.JAVA_OBJECT, Object.class },
				{ Types.LONGNVARCHAR, String.class },
				{ Types.LONGVARBINARY, Object.class },
				{ Types.NCHAR, String.class },
				{ Types.NCLOB, Object.class },
				{ Types.NULL, Object.class },
				{ Types.NUMERIC, Double.class },
				{ Types.NVARCHAR, String.class },
				{ Types.OTHER, Object.class },
				{ Types.REAL, Double.class },
				{ Types.REF, Object.class },
				{ Types.ROWID, Integer.class },
				{ Types.SMALLINT, Integer.class },
				{ Types.SQLXML, Object.class },
				{ Types.STRUCT, Object.class },
				{ Types.TIME, java.util.Date.class },
				{ Types.TIMESTAMP, java.util.Date.class },
				{ Types.TINYINT, Short.class },
				{ Types.VARBINARY, Object.class },
				{ Types.VARCHAR, String.class }
		};
		Tabular tabular = new TabularMock();
		
		for (Object[] datum : data) {
			Class<?> expected = (Class<?>)datum[1];
			Class<?> actual = tabular.classForSqlType((int) datum[0]);
			assertEquals(expected, actual);
		}
	}
	
	@Test
	public void test_printCsv_empTable() throws IOException, SQLException
	{
		TabularMock tabular = createEmpTable();
		
		try (
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				PrintWriter pw = new PrintWriter(new OutputStreamWriter(baos, StandardCharsets.UTF_8));
			)
		{
			tabular.printCsv(pw);
			pw.close();
			baos.close();
			String actual = baos.toString();
			assertEquals(  "EmpId,FirstName,LastName\n"
					     + "12345,Fred,Flintstone\n"
					     + "7654321,Barney,Rubble\n",
					     actual);
		}		
	}

	@Test
	public void test_printTable_empTable() throws IOException, SQLException
	{
		TabularMock tabular = createEmpTable();
		
		try (
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				PrintWriter pw = new PrintWriter(new OutputStreamWriter(baos, StandardCharsets.UTF_8));
			)
		{
			tabular.printTable(pw, "No rows returned.");
			pw.close();
			baos.close();
			String actual = baos.toString();
			assertEquals(  "+-------+---------+----------+\n"
					     + "| EmpId |FirstName| LastName |\n"
					     + "+-------+---------+----------+\n"
					     + "|12345  |Fred     |Flintstone|\n"
					     + "|7654321|Barney   |Rubble    |\n"
					     + "+-------+---------+----------+\n"
					     + "2 row(s) returned.\n",
					     actual);
		}
	}
	
	@Test
	public void test_repChar() {
		Tabular tabular = new TabularMock();
		
		assertEquals("", tabular.repChar(' ', 0));
		assertEquals("    ", tabular.repChar(' ', 4));
		assertEquals("###", tabular.repChar('#', 3));
		assertEquals("------", tabular.repChar('-', 6));
	}
	
	private TabularMock createEmpTable()
	{
		TabularMock tabular = new TabularMock();
		
		Column[] cols = { 
				new Column("EmpId", Long.class, 10),
				new Column("FirstName", String.class, 14),
				new Column("LastName", String.class, 14)
		};
		tabular.mock_setCols(cols);
		
		Row row = new Row(cols.length);
		row.setValue(1, Long.valueOf(12345));
		row.setValue(2, "Fred");
		row.setValue(3, "Flintstone");
		tabular.mock_addRow(row);
		
		row = new Row(cols.length);
		row.setValue(1, Long.valueOf(7654321));
		row.setValue(2, "Barney");
		row.setValue(3, "Rubble");
		tabular.mock_addRow(row);
		
		return tabular;
	}
}
